#include <OS.h>
#include <stdio.h>
#include <unistd.h>


inline const int
ffs_org(int value)
{
	if (!value)
		return 0;

	// ToDo: This can certainly be optimized (e.g. by binary search). Or not
	// unlikely there's a single assembler instruction...
	for (int i = 1; i <= (int)sizeof(value) * 8; i++, value >>= 1) {
		if (value & 1)
			return i;
	}

	// never gets here
	return 0;
}


inline const int
ffs_new(int value) 
{
	int result;
	if (!value)
		return 0; 
	__asm __volatile ("bsfl %1,%0" : "=r" (result) : "rm" (value));
	return ++result;
}


inline const int
ffs_alt(int value)
{
	static const int index[] = { 
		1, 2, 29, 3, 30, 15, 25, 4, 31, 23, 21, 16, 26, 18, 5, 9, 32,
		28, 14, 24, 22, 20, 17, 8, 27, 13, 19, 7, 12, 6, 11, 10 };

	if (value == 0)
		return 0;

	return index[(value & -value) * 125613361U >> 27];
#if 0
	switch( (value & -value) * 125613361U >>27) {
		case 0: return 1;
		case 1: return 2;
		case 2: return 29;
		case 3: return 3;
		case 4: return 30;
		case 5: return 15;
		case 6: return 25;
		case 7: return 4;
		case 8: return 31;
		case 9: return 23;
		case 10: return 21;
		case 11: return 16;
		case 12: return 26;
		case 13: return 18;
		case 14: return 5;
		case 15: return 9;
		case 16: return 32;
		case 17: return 28;
		case 18: return 14;
		case 19: return 24;
		case 20: return 22;
		case 21: return 20;
		case 22: return 17;
		case 23: return 8;
		case 24: return 27;
		case 25: return 13;
		case 26: return 19;
		case 27: return 7;
		case 28: return 12;
		case 29: return 6;
		case 30: return 11;
		case 31: return 10;
//		case 32: return ;
//		case 0: return ;
//		case 0: return ;
//		case 0: return ;
//		case 0: return ;
	}

	unsigned int w = value;
	w &= -w;
	w *= 125613361U;
	w >>= 27;
	return index[w];
#endif
}


int add_ffs_org(int value) {
	int result = 0;
	for (int i=0; i < 100000; i++)
		result += ffs_org(value);
	return result;
}


int add_ffs_alt(int value) {
	int result = 0;
	for (int i=0; i < 100000; i++)
		result += ffs_alt(value);
	return result;
}


int add_ffs_builtin(int value) {
	int result = 0;
	for (int i=0; i < 100000; i++)
		result += __builtin_ffs(value);
	return result;
}


int add_ffs_new(int value) {
	int result = 0;
	for (int i=0; i < 100000; i++)
		result += ffs_new(value);
	return result;
}


bigtime_t time_org(int value) {
	int result;
	bigtime_t time = real_time_clock_usecs();
	result = add_ffs_org(value);
	return real_time_clock_usecs() - time;
}

bigtime_t time_alt(int value) {
	int result;
	bigtime_t time = real_time_clock_usecs();
	result = add_ffs_alt(value);
	return real_time_clock_usecs() - time;
}

bigtime_t time_new(int value) {
	int result;
	bigtime_t time = real_time_clock_usecs();
	result = add_ffs_new(value);
	return real_time_clock_usecs() - time;
}

bigtime_t time_builtin(int value) {
	int result;
	bigtime_t time = real_time_clock_usecs();
	result = add_ffs_builtin(value);
	return real_time_clock_usecs() - time;
}

int
main(int argc, char *argv[])
{
	bigtime_t time;
    uint val, i;
    bool ok;	
		
	for (i = 0; i < 33; i++) {
		//correct = (correct + 1) % 
		val = 1 << i;
		if (i == 32) val = 0;
		
		ok = (ffs_org(val) == ffs_alt(val)) && (ffs_org(val) == ffs_new(val)) && (ffs_org(val) == __builtin_ffs(val));
		printf("%10u", val);
		printf("\t%3d %6d", ffs_org(val), time_org(val));
		printf("\t%3d %6d", ffs_alt(val), time_alt(val));
		printf("\t%3d %6d", ffs_new(val), time_new(val));
		printf("\t%3d %6d", __builtin_ffs(val), time_builtin(val));
		printf(" %s\n", ok ? "same" : "diffs");
	}	
}
